def get_default_searchspace(problem_type, num_classes=None):
    spaces = {
        # 'C': Real(lower=1e-4, upper=1e5, default=1),
        # 'tokenizer': Categorical('split', 'sentencepiece'),
        # 'fit_intercept': Categorical(True', False),
    }
    return spaces
